#! /usr/local/bin/perl

use strict;
use utf8;
use Encode;

# 環境変数LANGが*.UTF-8のときはUTF-8、それ以外はCP932でメッセージ表示します
my $os_enc = 'cp932';
if ($ENV{LANG} =~ m/.+\.UTF-8$/) {
	$os_enc = 'UTF-8';
}

binmode STDOUT, ":encoding($os_enc)";
binmode STDERR, ":encoding($os_enc)";

sub DEC { Encode::decode($os_enc, shift) }

my $in_file = '';
my $out_file = '';

sub usage()
{
	print "使用方法: escape_message.pl [オプション] メッセージファイル\n";
	print "機    能: メッセージファイル中のASCII以外の文字を\\uXXXX形式に変換します\n";
	print "オプション:\n";
	print "    -o|--output ファイル名  指定したファイルに変換したメッセージを保存します\n";
	print "                            (デフォルト: メッセージファイル + \".new\")\n";
	print "\n";
	print "    -?|--help               使用方法を表示します\n";
}

while (my $var = Encode::decode($os_enc, shift @ARGV)) {
	if ($var =~ /^-/) {
		if ($var eq '-?' or $var eq '--help') {
			usage();
			exit(1);
		} elsif ($var eq '-o' or $var eq '--output') {
			if ($#ARGV < 0) {
				printf("ERROR: オプション%sの後にファイル名がありません\n", $var);
				exit(1);
			}
			$out_file = Encode::decode($os_enc, shift @ARGV);
		} else {
			printf("ERROR: 不正なオプションです: %s\n", $var);
			exit(1);
		}
	} else {
		if ($in_file eq '') {
			$in_file = $var;
		} else {
			printf("ERROR: 不要なコマンドライン引数です: %s\n", $var);
			exit(1);
		}
	}
}

if ($in_file eq '') {
	print "ERROR: メッセージファイルを指定してください\n";
	exit(1);
}

if ($out_file eq '') {
	$out_file = $in_file . '.new';
}

open(my $IN, '<:unix:encoding(UTF-8)', $in_file) or die "ERROR: ファイルを開けません: $in_file: " . DEC($!) . "\n";
open(my $OUT, '>:unix:encoding(iso-8859-1)', $out_file) or die "ERROR: ファイルを開けません: $out_file: " . DEC($!) . "\n";

my $line = 0;
while (<$IN>) {
	$line++;

	my $modified = 0;
	my $orig = $_;
	while (m/([^\t\r\n -~])/) {
		my $from = $1;
		my $code = ord($1);
		my $to = '';
		if (0xD800 <= $code and $code <= 0xDFFF) {
			printf("WARNING: %d行目: 変換できない文字があります: %s\n", $line, $from);
		} elsif ($code <= 0xFFFF) {
			$to = sprintf("\\u%04X", $code);
		} elsif ($code <= 0x10FFFF) {
			# サロゲートペア
			$code -= 0x10000;
			my $code1 = $code / 0x400 + 0xD800;
			my $code2 = $code % 0x400 + 0xDC00;
			$to = sprintf("\\u%04X\\u%04X", $code1, $code2);
		} else {
			printf("WARNING: %d行目: 変換できない文字があります: %s\n", $line, $from);
		}
		s/$from/$to/g;
		$modified = 1;
	}

	print $OUT $_;

	if ($modified != 0) {
		printf("%4.1d行目: 変更前: %s          変更後: %s", $line, $orig, $_);
	}
}

close($OUT);
close($IN);

