#! /usr/local/bin/perl

use strict;
use utf8;
use Encode;

# 環境変数LANGが*.UTF-8のときはUTF-8、それ以外はCP932でメッセージ表示します
my $os_enc = 'cp932';
if ($ENV{LANG} =~ m/.+\.UTF-8$/) {
	$os_enc = 'UTF-8';
}

binmode STDOUT, ":encoding($os_enc)";
binmode STDERR, ":encoding($os_enc)";

sub DEC { Encode::decode($os_enc, shift) }

my $text_file = 'script_text.txt';
my $in_dir = '';
my $script_file = 'skript.csv';
my $choice_file = 'choiceliste.csv';

my $script_encoding = 'iso-8859-1';

my $out_dir = 'SCRIPT_NEW';
my $out_script_file = $script_file;
my $out_choice_file = $choice_file;

sub usage()
{
	print "使用方法: replace_script_text.pl [オプション]\n";
	print "機    能: skript.csv, choiceliste.csvのテキストを翻訳ファイルのものに変更します\n";
	print "オプション:\n";
	print "       --latin1             skript.csvをLatin-1(iso-8859-1)で読み込みます\n";
	print "                            英語のファイルを使用する場合に指定してください\n";
	print "\n";
	print "    -t|--text   ファイル名  指定したファイルから翻訳テキストを読み込みます\n";
	print "                            (デフォルト: script_text.txt)\n";
	print "    -i|--indir  ディレクトリ名\n";
	print "                            指定したディレクトリからスクリプトを読み込みます\n";
	print "    -o|--outdir ディレクトリ名\n";
	print "                            指定したディレクトリにテキストを変更したファイルを\n";
	print "                            作成します\n";
	print "                            (デフォルト: SCRIPT_NEW)\n";
	print "    -?|--help               使用方法を表示します\n";
}

while (my $var = Encode::decode($os_enc, shift @ARGV)) {
	if ($var =~ /^-/) {
		if ($var eq '-?' or $var eq '--help') {
			usage();
			exit(1);
		} elsif ($var eq '--latin1') {
			$script_encoding = 'iso-8859-1';
		} elsif ($var eq '-t' or $var eq '--text') {
			if ($#ARGV < 0) {
				printf("ERROR: オプション%sの後にファイル名がありません\n", $var);
				exit(1);
			}
			$text_file = Encode::decode($os_enc, shift @ARGV);
		} elsif ($var eq '-i' or $var eq '--indir') {
			if ($#ARGV < 0) {
				printf("ERROR: オプション%sの後にディレクトリ名がありません\n", $var);
				exit(1);
			}
			$in_dir = Encode::decode($os_enc, shift @ARGV);
		} elsif ($var eq '-o' or $var eq '--outdir') {
			if ($#ARGV < 0) {
				printf("ERROR: オプション%sの後にディレクトリ名がありません\n", $var);
				exit(1);
			}
			$out_dir = Encode::decode($os_enc, shift @ARGV);
		} else {
			printf("ERROR: 不正なオプションです: %s\n", $var);
			exit(1);
		}
	} else {
		printf("ERROR: 不要なコマンドライン引数です: %s\n", $var);
		exit(1);
	}
}

if ($in_dir ne '') {
	$script_file = "$in_dir/$script_file";
	$choice_file = "$in_dir/$choice_file";
}

if ($out_dir ne '') {
	$out_script_file = "$out_dir/$out_script_file";
	$out_choice_file = "$out_dir/$out_choice_file";
}

my %script_text;
my %ja_text;
my %ja_text2;

sub add_ja_text($$)
{
	my ($text, $ja_text) = @_;
	return if ($text eq '');
	return if ($ja_text eq '');
	return if ($ja_text eq '0');
	return if ($ja_text eq '機械翻訳');
	return if ($ja_text eq '原文');

	if (not exists $ja_text{$text}) {
		$ja_text{$text} = $ja_text;
	}
}

sub add_ja_text2($$)
{
	my ($text, $ja_text) = @_;
	return if ($text eq '');
	return if ($ja_text eq '');
	return if ($ja_text eq '0');
	return if ($ja_text eq '機械翻訳');
	return if ($ja_text eq '原文');

	if (not exists $ja_text2{$text}) {
		$ja_text2{$text} = $ja_text;
	}
}

sub add_script_text($$$$$$$$)
{
	my ($line, $type, $id1, $id2, $func, $text, $ja_text, $ja_text2) = @_;
	my $key = "$type\t$id1\t$id2";
	if (exists $script_text{$key}) {
		my ($prev_line, $prev_func, $prev_text, $prev_ja_text, $prev_ja_text2) = split("\t", $script_text{$key}, 5);
		printf("WARNING: 同じ番号のテキストが存在します: %d行目, %d行目: %s %s\n", $prev_line, $line, $id1, $id2);
	} else {
		$script_text{$key} = join("\t", $line, $func, $text, $ja_text, $ja_text2);
	}

	add_ja_text($text, $ja_text);
	add_ja_text2($text, $ja_text2);
}

sub tune_ja_text()
{
	# 原文前後の空白文字を除去したキーで重複登録する
	my @keys = keys %ja_text;
	foreach my $text ( @keys ) {
		if ($text =~ m/^[ \t]+/ or $text =~ m/[ \t]+$/) {
			my $text2 = $text;
			$text2 =~ s/^[ \t]+//;
			$text2 =~ s/[ \t]+$//;

			if (not exists $ja_text{$text2}) {
				$ja_text{$text2} = $ja_text{$text};
			}
		}
	}

	@keys = keys %ja_text2;
	foreach my $text ( @keys ) {
		if ($text =~ m/^[ \t]+/ or $text =~ m/[ \t]+$/) {
			my $text2 = $text;
			$text2 =~ s/^[ \t]+//;
			$text2 =~ s/[ \t]+$//;

			if (not exists $ja_text2{$text2}) {
				$ja_text2{$text2} = $ja_text2{$text};
			}
		}
	}
}

sub get_script_text($$$$$$)
{
	my ($type, $line, $id1, $id2, $func, $text) = @_;
	my $key = "$type\t$id1\t$id2";

	if (not exists $script_text{$key}) {
		# 翻訳ファイルに存在しないテキスト
		return '';
	}

	my ($found_line, $found_func, $found_text, $ja_text, $ja_text2) = split("\t", $script_text{$key}, 5);

	if ($text ne $found_text) {
		printf("WARNING: 翻訳ファイルとスクリプトファイルの原文が違います\n");
		printf("\tスクリプト:%8.1d行目\t%s\n", $line, $text);
		printf("\t翻訳      :%8.1d行目\t%s\n", $found_line, $found_text);
	}

	if ($func ne $found_func) {
		if ($type eq 'script') {
			printf("WARNING: 翻訳ファイルとスクリプトファイルの関数が違います\n");
		} else {
			printf("WARNING: 翻訳ファイルとスクリプトファイルのスクリプトIDが違います\n");
		}
		printf("\tスクリプト:%8.1d行目\t%s\n", $line, $func);
		printf("\t翻訳      :%8.1d行目\t%s\n", $found_line, $found_func);
	}

	if ($ja_text eq '') {
		return $ja_text2;
	} elsif ($ja_text eq '原文') {
		return '';
	} elsif ($ja_text eq '機械翻訳') {
		return $ja_text2;
	} elsif ($ja_text ne '0') {
		return $ja_text;
	}

	if ($type eq 'choice') {
		# 選択肢に対応するスクリプトの原文が同じとき、その翻訳があればそれを使用する
		my $script_key = "script\t$func\t1";

		if (exists $script_text{$script_key}) {
			my ($script_line, $script_func, $script_text, $script_ja_text, $script_ja_text2) = split("\t", $script_text{$script_key}, 5);
			if ($script_text eq $text) {
				if ($script_ja_text eq '') {
					if ($script_ja_text2 ne '') {
						return $script_ja_text2;
					}
				} elsif ($script_ja_text eq '原文') {
					return '';
				} elsif ($script_ja_text eq '機械翻訳') {
					if ($script_ja_text2 ne '') {
						return $script_ja_text2;
					}
				} elsif ($script_ja_text ne '0') {
					return $script_ja_text;
				}
			}
		}
	}

	if (exists $ja_text{$text}) {
		return $ja_text{$text};
	}

	if (exists $ja_text2{$text}) {
		return $ja_text2{$text};
	}

	if ($text =~ m/^[ \t]+/ or $text =~ m/[ \t]+$/) {
		# 原文前後の空白文字を除去して翻訳を探す
		$text =~ s/^[ \t]+//;
		$text =~ s/[ \t]+$//;

		if (exists $ja_text{$text}) {
			return $ja_text{$text};
		}

		if (exists $ja_text2{$text}) {
			return $ja_text2{$text};
		}
	}

	return '';
}

my $num_script_replaced = 0;
my $num_choice_replaced = 0;

print "翻訳ファイル読み込み中...\n";
open(my $TEXT, '<:unix:encoding(UTF-8)', $text_file) or die "ERROR: ファイルを開けません: $text_file: " . DEC($!) . "\n";
my $line = 0;
while (<$TEXT>) {
	$line++;

	s/[\r\n]+//;
	next if (m/^#/);
	next if (m/^$/);

	my @work = split("\t", $_);
	if ($#work < 4) {
		printf("ERROR: 翻訳テキストの項目数が不足しています: %d行目\n", $line);
		exit(1);
	}

	my $type = $work[0];
	my $id1 = $work[1];
	my $id2 = $work[2];
	my $func = $work[3];
	my $text = $work[4];
	my $ja_text = '';
	my $ja_text2 = '';
	if (5 <= $#work) {
		$ja_text = $work[5];
	}
	if (6 <= $#work) {
		$ja_text2 = $work[6];
	}

	next if ($ja_text eq '' and $ja_text2 eq '');

	if ($type eq 'script' or $type eq 'choice') {
		add_script_text($line, $type, $id1, $id2, $func, $text, $ja_text, $ja_text2);
	} else {
		printf("ERROR: 不正な翻訳テキストです: %d行目\n", $line);
		exit(1);
	}
}
close($TEXT);

tune_ja_text();

if (not -d $out_dir) {
	print "出力ディレクトリ作成\n";
	mkdir($out_dir) or die "ERROR: ディレクトリを作成できません: $out_dir: " . DEC($!) . "\n";
}

print "スクリプトファイル置換中...\n";
open(my $IN, "<:unix:encoding($script_encoding)", $script_file) or die "ERROR: ファイルを開けません: $script_file: " .DEC($!) . "\n";
open(my $OUT, '>:unix:encoding(UTF-8)', $out_script_file) or die "ERROR: ファイルを開けません: $out_script_file: " . DEC($!) . "\n";
$line = 0;
while (<$IN>) {
	$line++;
	if (m/^([0-9]+);([0-9]+);(if(?:item)?active\([0-9]+\),)?(say|think)\((.+)\)(;+[^;]*[\r\n]+)$/) {
		my ($id1, $id2, $ifactive, $func, $text, $crlf) = ($1, $2, $3, $4, $5, $6);
		my $ja_text = get_script_text('script', $line, $id1, $id2, $func, $text);
		if ($ja_text eq '') {
			print $OUT $_;
		} else {
			if ($ja_text =~ m/^<#>/) {
				$ja_text =~ s/^<#>/$id1-$id2/;
			}
			print $OUT $id1, ";", $id2, ";", $ifactive, $func, "(", $ja_text, ")", $crlf;
			$num_script_replaced++;
		}
	} elsif (m/^([0-9]+);([0-9]+);(if(?:item)?active\([0-9]+\),)?(saynscp?)\(([0-9]+),(.+)\)(;+[^;]*[\r\n]+)$/) {
		my ($id1, $id2, $ifactive, $func, $param, $text, $crlf) = ($1, $2, $3, $4, $5, $6, $7);
		my $ja_text = get_script_text('script', $line, $id1, $id2, $func, $text);
		if ($ja_text eq '') {
			print $OUT $_;
		} else {
			if ($ja_text =~ m/^<#>/) {
				$ja_text =~ s/^<#>/$id1-$id2/;
			}
			print $OUT $id1, ";", $id2, ";", $ifactive, $func, "(", $param, ",", $ja_text, ")", $crlf;
			$num_script_replaced++;
		}
	} elsif (m/^([0-9]+);([0-9]+);(if(?:item)?active\([0-9]+\),)?(saysound)\(([0-9]+,[0-9]+,|char,)?(.+)\)(;+[^;]*[\r\n]+)$/) {
		my ($id1, $id2, $ifactive, $func, $param, $text, $crlf) = ($1, $2, $3, $4, $5, $6, $7);
		my $ja_text = get_script_text('script', $line, $id1, $id2, $func, $text);
		if ($ja_text eq '') {
			print $OUT $_;
		} else {
			if ($ja_text =~ m/^<#>/) {
				$ja_text =~ s/^<#>/$id1-$id2/;
			}
			print $OUT $id1, ";", $id2, ";", $ifactive, $func, "(", $param, $ja_text, ")", $crlf;
			$num_script_replaced++;
		}
	} elsif (m/^([0-9]+);([0-9]+);(if(?:item)?active\([0-9]+\),)?(saysoundfile)\(((?:[0-9]+,[0-9]+,[^,]+|char(?:,[^,]+)?)),(.+)\)(;+[^;]*[\r\n]+)$/) {
		my ($id1, $id2, $ifactive, $func, $param, $text, $crlf) = ($1, $2, $3, $4, $5, $6, $7);
		my $ja_text = get_script_text('script', $line, $id1, $id2, $func, $text);
		if ($ja_text eq '') {
			print $OUT $_;
		} else {
			if ($ja_text =~ m/^<#>/) {
				$ja_text =~ s/^<#>/$id1-$id2/;
			}
			print $OUT $id1, ";", $id2, ";", $ifactive, $func, "(", $param, ",", $ja_text, ")", $crlf;
			$num_script_replaced++;
		}
	} else {
		print $OUT $_;
	}
}
close($OUT);
close($IN);

print "選択肢ファイル置換中...\n";
open(my $IN, '<:unix:encoding(UTF-8)', $choice_file) or die "ERROR: ファイルを開けません: $choice_file: " . DEC($!) . "\n";
open(my $OUT, '>:unix:encoding(UTF-8)', $out_choice_file) or die "ERROR: ファイルを開けません: $out_choice_file: " . DEC($!) . "\n";
$line = 0;
while (<$IN>) {
	$line++;
	if (m/^([0-9]+);([0-9]+);([^;]+);(.+);([0-9]+)([ \t]*[\r\n]+)$/) {
		my ($id1, $id2, $param, $text, $script_id, $crlf) = ($1, $2, $3, $4, $5, $6);
		my $ja_text = get_script_text('choice', $line, $id1, $id2, $script_id, $text);
		if ($ja_text eq '') {
			print $OUT $_;
		} else {
			if ($ja_text =~ m/^<#>/) {
				$ja_text =~ s/^<#>/$id1-$id2/;
			}
			print $OUT $id1, ";", $id2, ";", $param, ";", $ja_text, ";", $script_id, $crlf;
			$num_choice_replaced++;
		}
	} else {
		print $OUT $_;
	}
}
close($OUT);
close($IN);

print "完了\n\n";

printf("スクリプト: %d件置換\n", $num_script_replaced);
printf("選択肢    : %d件置換\n", $num_choice_replaced);

