#! /usr/local/bin/perl

use strict;
use utf8;
use Encode;

# 環境変数LANGが*.UTF-8のときはUTF-8、それ以外はCP932でメッセージ表示します
my $os_enc = 'cp932';
if ($ENV{LANG} =~ m/.+\.UTF-8$/) {
	$os_enc = 'UTF-8';
}

binmode STDOUT, ":encoding($os_enc)";
binmode STDERR, ":encoding($os_enc)";

sub DEC { Encode::decode($os_enc, shift) }

my $in_file = '';
my $out_file = '';

sub usage()
{
	print "使用方法: unescape_message.pl [オプション] メッセージファイル\n";
	print "機    能: メッセージファイル中の\\uXXXXを指定された文字に変換します\n";
	print "オプション:\n";
	print "    -o|--output ファイル名  指定したファイルに変換したメッセージを保存します\n";
	print "                            (デフォルト: メッセージファイル + \".txt\")\n";
	print "\n";
	print "    -?|--help               使用方法を表示します\n";
}

while (my $var = Encode::decode($os_enc, shift @ARGV)) {
	if ($var =~ /^-/) {
		if ($var eq '-?' or $var eq '--help') {
			usage();
			exit(1);
		} elsif ($var eq '-o' or $var eq '--output') {
			if ($#ARGV < 0) {
				printf("ERROR: オプション%sの後にファイル名がありません\n", $var);
				exit(1);
			}
			$out_file = Encode::decode($os_enc, shift @ARGV);
		} else {
			printf("ERROR: 不正なオプションです: %s\n", $var);
			exit(1);
		}
	} else {
		if ($in_file eq '') {
			$in_file = $var;
		} else {
			printf("ERROR: 不要なコマンドライン引数です: %s\n", $var);
			exit(1);
		}
	}
}

if ($in_file eq '') {
	print "ERROR: メッセージファイルを指定してください\n";
	exit(1);
}

if ($out_file eq '') {
	$out_file = $in_file . '.txt';
}

open(my $IN, '<:unix:encoding(iso-8859-1)', $in_file) or die "ERROR: ファイルを開けません: $in_file: " . DEC($!) . "\n";
open(my $OUT, '>:unix:encoding(UTF-8)', $out_file) or die "ERROR: ファイルを開けません: $out_file: " . DEC($!) . "\n";

my $line = 0;
while (<$IN>) {
	$line++;

	my $modified = 0;
	my $orig = $_;

	# サロゲートペア
	while (m/(\\u(D[89abAB][0-9a-fA-F]{2})\\u(D[c-fC-F][0-9a-fA-F]{2}))/) {
		my $from = quotemeta($1);
		my $code1 = hex($2);
		my $code2 = hex($3);
		my $to = chr(0x10000 + ($code1 - 0xD800) * 0x400 + ($code2 - 0xDC00));
		s/$from/$to/g;
		$modified = 1;
	}

	# 念のためサロゲートペアの片割れを除去
	#s/\\uD[89a-FA-F][0-9a-fA-F]{2}//g;
	while (m/(\\uD[89a-fA-F][0-9a-fA-F]{2})/) {
		printf("WARNING: %d行目: 変換できない文字があります: %s\n", $line, $1);
		my $from = quotemeta($1);
		s/$from//g;
	}

	while (m/(\\u([0-9a-fA-F]{4}))/) {
		my $from = quotemeta($1);
		my $to = chr(hex("0x$2"));
		s/$from/$to/g;
		$modified = 1;
	}

	print $OUT $_;

	if ($modified != 0) {
		printf("%4.1d行目: 変更前: %s          変更後: %s", $line, $orig, $_);
	}
}

close($OUT);
close($IN);

